/* EntryFileDialog.java
 *
 * created: Mon Dec  7 1998
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998-2003  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/pathsoft/repository/powmap/diana/components/EntryFileDialog.java,v 1.31 2003/03/07 09:23:09 kmr Exp $
 */

package diana.components;

import diana.Options;
import uk.ac.sanger.pathogens.*;
import uk.ac.sanger.pathogens.embl.Entry;
import uk.ac.sanger.pathogens.embl.DocumentEntryFactory;
import uk.ac.sanger.pathogens.embl.ReadFormatException;
import uk.ac.sanger.pathogens.embl.EntryInformation;
import uk.ac.sanger.pathogens.embl.EntryInformationException;
import uk.ac.sanger.pathogens.embl.InvalidRelationException;
import uk.ac.sanger.pathogens.embl.InvalidKeyException;
import uk.ac.sanger.pathogens.embl.InvalidQualifierException;
import uk.ac.sanger.pathogens.embl.SimpleEntryInformation;
import uk.ac.sanger.pathogens.embl.Key;
import uk.ac.sanger.pathogens.embl.Qualifier;
import uk.ac.sanger.pathogens.embl.QualifierInfo;
import uk.ac.sanger.pathogens.embl.QualifierInfoException;

import java.awt.*;
import java.io.*;

import javax.swing.*;

/**
 *  This class is a FileDialog that can read EMBL Entry objects.
 *
 *  @author Kim Rutherford
 *  @version $Id: EntryFileDialog.java,v 1.31 2003/03/07 09:23:09 kmr Exp $
 **/

public class EntryFileDialog extends JFileChooser {
  /**
   *  Create a new EntryFileDialog component.
   *  @param owner The component where this dialog was created.
   **/
  public EntryFileDialog (final JFrame owner) {
    super ();
    this.owner = owner;

    if (last_directory == null) {
      if (Options.getOptions ().getProperty ("default_directory") != null) {
        final String default_directory =
          Options.getOptions().getProperty("default_directory");
        setCurrentDirectory (new File (default_directory));
      } else {
        if (last_directory == null) {
          System.err.println ("user.dir: " + System.getProperty ("user.dir"));
          last_directory = new File (System.getProperty ("user.dir"));
        }
      }
    } else {
      setCurrentDirectory (last_directory);
    }

    setCurrentDirectory (last_directory);
    setFileSelectionMode (JFileChooser.FILES_AND_DIRECTORIES);

    final javax.swing.filechooser.FileFilter artemis_filter =
      new javax.swing.filechooser.FileFilter () {
        public boolean accept (final File file) {
          if (file.getName ().endsWith (".dna") ||
              file.getName ().endsWith (".seq") ||
              file.getName ().endsWith (".tab") ||
              file.getName ().endsWith (".embl") ||
              file.getName ().endsWith (".tab_embl") ||
              file.getName ().endsWith (".gff") ||
              file.getName ().endsWith (".feature") ||
              file.getName ().endsWith (".feat") ||
              file.isDirectory ()) {
            return true;
          } else {
            return false;
          }
        }

        public String getDescription () {
          return "Artemis files";
        }
      };

    final javax.swing.filechooser.FileFilter feature_filter =
      new javax.swing.filechooser.FileFilter () {
        public boolean accept (final File file) {
          if (file.getName ().endsWith (".tab") ||
              file.getName ().endsWith (".embl") ||
              file.getName ().endsWith (".tab_embl") ||
              file.getName ().endsWith (".gff") ||
              file.getName ().endsWith (".feature") ||
              file.getName ().endsWith (".feat") ||
              file.isDirectory ()) {
            return true;
          } else {
            return false;
          }
        }

        public String getDescription () {
          return "Feature files";
        }
      };

    final javax.swing.filechooser.FileFilter sequence_filter =
      new javax.swing.filechooser.FileFilter () {
        public boolean accept (final File file) {
          if (file.getName ().endsWith (".dna") ||
              file.getName ().endsWith (".embl") ||
              file.getName ().endsWith (".tab_embl") ||
              file.getName ().endsWith (".seq") ||
              file.isDirectory ()) {
            return true;
          } else {
            return false;
          }
        }

        public String getDescription () {
          return "Sequence files";
        }
      };

    addChoosableFileFilter (sequence_filter);
    addChoosableFileFilter (feature_filter);
    addChoosableFileFilter (artemis_filter);
  }

  /**
   *  Used to remember the directory the FileDialog was in when the user
   *  pressed OK.  This is used as the starting directory next time.
   **/
  private static File last_directory = null;

  /**
   *  Return an uk.ac.sanger.pathogens.embl.Entry object representing the file
   *  the user has selected with the dialog or null if the read failed for any
   *  reason.
   *  @param entry_information The EntryInformation to use when reading.  This
   *    supplies the list of valid keys and qualifiers.  If a key or qualifier
   *    is read that is incompatible with this EntryInformation object the
   *    EntryInformation will be changed to cope.
   *  @param listener The object to which InputStreamProgressEvents will be
   *    send while reading.
   *  @exception EntryInformationException Thrown if an Entry using the given
   *    EntryInformation object cannot contain the Key, Qualifier or
   *    Key/Qualifier combination of one of the features in the Document.
   *  @param show_progress If true a InputStreamProgressDialog will be shown
   *    while loading.
   **/
  public Entry getEntry (final EntryInformation entry_information,
                         final InputStreamProgressListener listener,
                         final boolean show_progress) {
    showOpenDialog (owner);
    setDialogTitle ("Select a file ...");
    setDialogType (JFileChooser.OPEN_DIALOG);

    if (getSelectedFile () == null) {
      return null;
    }

    last_directory = getCurrentDirectory ();

    final File file =
      new File (getCurrentDirectory (), getSelectedFile ().getName ());

    System.err.println ("file: " + file);
    System.err.println ("current: " + getCurrentDirectory ());

    return getEntryFromFile (owner, file, entry_information, listener,
                             show_progress);
  }

  /**
   *  This exists only to get around a bug in the 1.1/1.2 code generator,
   *  which generates unverifiable code.
   *  @param frame Used when creating MessageDialog components.
   *  @param entry_file The file to read the Entry from.
   *  @param file_document The Document to read the Entry from (should be made
   *    from entry_file).
   *  @param entry_information The EntryInformation to use when reading.  This
   *    supplies the list of valid keys and qualifiers.  If a key or qualifier
   *    is read that is incompatible with this EntryInformation object the
   *    EntryInformation will be changed to cope.
   *  @param errors One String is added to this for each error.
   **/
  private static Entry
    getEntryFromFileHelper (final JFrame frame,
                            final File entry_file,
                            final FileDocument file_document,
                            final EntryInformation entry_information,
                            final StringVector errors)
      throws ReadFormatException, IOException {
    final int MAX_ERROR_COUNT = 100;

    while (errors.size () < MAX_ERROR_COUNT) {
      try {
        if (entry_file != null) {
          final Entry new_entry =
            DocumentEntryFactory.makeDocumentEntry (entry_information,
                                                    file_document);

          return new_entry;
        } else {
          new MessageDialog (frame,
                             "creating new entry: " + entry_file.getName ());

          final Entry new_entry =
            DocumentEntryFactory.makeDocumentEntry (file_document);

          return new_entry;
        }
      } catch (EntryInformationException e) {
        final String message = e.getMessage ();

        errors.add (message);

        if (errors.size () == 1) {
          final YesNoDialog yes_no_dialog =
            new YesNoDialog (frame,
                             message + " - ignore error and continue?");

          if (!yes_no_dialog.getResult ()) {
            return null;
          }
        }

        handleOpenException (entry_information, e);
      }
    }

    new MessageDialog (frame,
                       "aborted - too many errors while reading: " +
                       entry_file.getName ());

    return null;
  }

  /**
   *  Read and return an Entry from the given File.
   *  @param frame Used when creating MessageDialog components.
   *  @param entry_file The file to read the Entry from.
   *  @param entry_information The EntryInformation to use when reading.  This
   *    supplies the list of valid keys and qualifiers.  If a key or qualifier
   *    is read that is incompatible with this EntryInformation object the
   *    EntryInformation will be changed to cope.
   *  @param listener The object to which InputStreamProgressEvents will be
   *    send while reading.
   *  @param show_progress If true a InputStreamProgressDialog will be shown
   *    while loading.
   **/
  public static Entry
    getEntryFromFile (final JFrame frame,
                      final File entry_file,
                      final EntryInformation entry_information,
                      final InputStreamProgressListener listener,
                      final boolean show_progress) {
    StringVector errors = new StringVector ();

    final InputStreamProgressDialog progress_dialog;

    if (show_progress) {
      progress_dialog =
        new InputStreamProgressDialog (frame, "Reading ...",
                                       "Reading from " +
                                       entry_file.getName (), false);
    } else {
      progress_dialog = null;
    }

    try {
      final InputStreamProgressListenerVector listeners =
        new InputStreamProgressListenerVector ();

      if (listener != null) {
        listeners.add (listener);
      }

      if (progress_dialog != null) {
        listeners.add (progress_dialog.getInputStreamProgressListener ());
      }

      final FileDocument file_document =
        new FileProgressDocument (entry_file, listeners);

      return getEntryFromFileHelper (frame, entry_file, file_document,
                                     entry_information, errors);
    } catch (ReadFormatException e) {
      final String message =
        "failed to open " + entry_file.getName () + ": " + e.getMessage () +
        " at line: " + e.getLineNumber ();
      System.out.println (message);
      new MessageDialog (frame, message);
    } catch (FileNotFoundException e) {
      final String message =
        "failed to open " + entry_file.getName () + ": file not found";
      new MessageDialog (frame, message);
    } catch (IOException e) {
      final String message =
        "failed to open " + entry_file.getName () + ": " + e.getMessage ();
      new MessageDialog (frame, message);
    } finally {
      if (errors.size () > 1) {
        final YesNoDialog dialog =
          new YesNoDialog (frame,
                           "There were warnings while reading.  " +
                           "View them now?");

        if (dialog.getResult ()) {
          final FileViewer viewer =
            new FileViewer ("Read errors from " + entry_file.getName ());

          final StringBuffer buffer = new StringBuffer ();

          for (int i = 0 ; i < errors.size () ; ++i) {
            buffer.append (errors.elementAt (i) + "\n");
          }

          viewer.clear ();
          viewer.appendString (buffer.toString ());
        }
      }

      if (progress_dialog !=null) {
        progress_dialog.dispose ();
      }
    }

    return null;
  }

  /**
   *  Save the given entry, prompting for a file name if necessary.
   *  @param include_diana_extensions If true then the any diana additions to
   *    the embl file format will be included in the output, otherwise they
   *    will be removed.  Also possible problems that would cause an entry to
   *    bounce from the EMBL submission process will be flagged if this is
   *    true.
   *  @param ask_for_name If true then always prompt for a new filename,
   *    otherwise prompt only when the entry name is not set.
   *  @param keep_new_name If ask_for_name is true a file will be written with
   *    the new name the user selects - if keep_new_name is true as well, then
   *    the entry will have it's name set to the new name, otherwise it will
   *    be used for this save and then discarded.
   *  @param destination_type Should be one of EMBL_FORMAT, GENBANK_FORMAT or
   *    ANY_FORMAT.  If ANY_FORMAT then the Entry will be saved in the
   *    same format it was created, otherwise it will be saved in the given
   *    format.
   **/
  void saveEntry (final diana.Entry entry,
                  final boolean include_diana_extensions,
                  final boolean ask_for_name,
                  final boolean keep_new_name,
                  final int destination_type) {

    try {
      if (ask_for_name || entry.getName () == null) {
        setDialogTitle ("Save to ...");
        setDialogType (JFileChooser.SAVE_DIALOG);
        showSaveDialog (owner);

        last_directory = getCurrentDirectory ();

        if (getSelectedFile () == null) {
          return;
        }

        final File file =
          new File (getCurrentDirectory (), getSelectedFile ().getName ());

        if (file.exists ()) {
          final YesNoDialog yes_no_dialog =
            new YesNoDialog (owner, "this file exists: " + file.getName () +
                             " overwrite it?");

          if (yes_no_dialog.getResult ()) {
            // yes - continue
          } else {
            // no
            return;
          }
        }

        final MessageDialog message =
          new MessageDialog (owner,
                             "saving to " + file.getName () + " ...",
                             false);

        try {
          if (include_diana_extensions) {
            entry.save (file, destination_type, false);
          } else {
            entry.saveStandardOnly (file, destination_type, true);
          }
        } catch (EntryInformationException e) {
          final YesNoDialog yes_no_dialog =
            new YesNoDialog (owner, "destination format can't handle all " +
                             "keys/qualifiers - continue?");
          if (yes_no_dialog.getResult ()) {
            try {
              if (include_diana_extensions) {
                entry.save (file, destination_type, true);
              } else {
                entry.saveStandardOnly (file, destination_type, true);
              }
            } catch (EntryInformationException e2) {
              throw new Error ("internal error - unexpected exception: "+ e);
            }
          } else {
            return;
          }
        }

        if (keep_new_name) {
          entry.setName (file.getName ());
        }

        if (message != null) {
          message.dispose ();
        }
      } else {
        final MessageDialog message =
          new MessageDialog (owner,
                             "saving to " + entry.getName () + " ...",
                             false);

        if (include_diana_extensions) {
          entry.save (destination_type);
        } else {
          entry.saveStandardOnly (destination_type);
        }

        message.dispose ();
      }
    } catch (ReadOnlyException e) {
      new MessageDialog (owner, "this entry is read only");
      return;
    } catch (IOException e) {
      new MessageDialog (owner, "error while saving: " + e);
      return;
    } catch (EntryInformationException e) {
      new MessageDialog (owner, "error while saving: " + e);
      return;
    }
  }

  /**
   *  Fix entry_information so that the given exception won't happen again.
   **/
  public static void
    handleOpenException (final EntryInformation entry_information,
                         final EntryInformationException exception) {
    if (exception instanceof InvalidKeyException) {
      final InvalidKeyException key_exception =
        (InvalidKeyException) exception;

      entry_information.addKey (key_exception.getKey ());
    }

    if (exception instanceof InvalidRelationException) {
      final InvalidRelationException relation_exception =
        (InvalidRelationException) exception;

      if (relation_exception.getQualifier () == null) {
        final Key key_to_add = relation_exception.getKey ();

        entry_information.addKey (key_to_add);
      } else {
        final Qualifier qualifier = relation_exception.getQualifier ();

        // change the QualifierInfo object in the EntryInformation so that it
        // can handle this key/qualifier combination

        final QualifierInfo qualifier_info =
          entry_information.getQualifierInfo (qualifier.getName ());

        final QualifierInfo new_qualifier_info;

        if (qualifier_info == null) {
          new_qualifier_info =
            new QualifierInfo (qualifier.getName (),
                               QualifierInfo.UNKNOWN, null, null, false);
        } else {
          new_qualifier_info =
            new QualifierInfo (qualifier_info.getName (),
                               qualifier_info.getType (), null, null, false);
        }

        try {
          entry_information.addQualifierInfo (new_qualifier_info);
        } catch (QualifierInfoException e) {
          throw new Error ("internal error - unexpected exception: " + e);
        }
      }
    }

    if (exception instanceof InvalidQualifierException) {
      final String exception_qualifier_name =
        ((InvalidQualifierException) exception).getQualifier ().getName ();

      // add it again, but make it less restrictive
      final QualifierInfo new_info =
        new QualifierInfo (exception_qualifier_name, QualifierInfo.UNKNOWN,
                           null, null, false);

      try {
        entry_information.addQualifierInfo (new_info);
      } catch (QualifierInfoException e) {
        throw new Error ("internal error - unexpected exception: " + e);
      }
    }
  }

  /**
   *  The JFrame reference that was passed to the constructor.
   **/
  private JFrame owner = null;
}
